import java.awt.Color;
import java.awt.GradientPaint;
import java.awt.Point;

/**
 * Eine Wand-, Boden- oder Deckenkachel des Labyrinths
 * oder eine Marke auf dem Labyrinthboden.
 * 
 * @author Bernd Gramlich
 * @version 2013-01-09
 */
public abstract class Kachel
{
    /*
     * Bild der Kachel ist ein Trapez.
     */
    protected Vieleck bild;
    
    /*
     * Die Faerbung der Kachel ist eine sich in
     * die Ferne abschwaechende Farbschattierung.
     */
    protected Color faerbung;
    
    /*
     * Die Koordinaten der Kachel relativ zum Betrachter.
     * 
     * Die Laengeneinheit im Koordinatensystem ist 2. Dadurch
     * haben Boden- und Deckenkacheln geradzahlige Koordinaten
     * (und implizit die ungeradzahlige z-Koordinate -1 bzw. 1).
     * Wandkacheln links und rechts haben ungeradzahlige
     * x-Koordinate und geradzahlige y-Koordinate. Wandkacheln
     * hinten haben geradzahlige x- und ungeradzahlige
     * y-Koordinate.
     */
    protected int positionX;
    protected int positionY;
    
    /*
     * Der Abstand des Kachelmittelpunkts vom Betrachter
     * (in Summenmetrik).
     */
    protected int abstand;
    
    /*
     * Der euklidische Abstand des Kachelmittelpunkts vom
     * Betrachter.
     */
    protected double euklidischerAbstand;
    
    /*
     * Die Hellsichtweite des Labyrinths.
     */
    protected double hellsichtweite;
    
    /*
     * Zwei Punkte auf der Kachel und ihre euklidischen
     * Abstaende vom Betrachter sowie die Zyklizitaet. Diese
     * Informationen werden gebraucht, um die Kachel mit
     * GradientPaint zwar nicht euklidisch, aber in
     * orthogonal-linearer Naeherung zu faerben.
     */
    protected Point P, Q;
    protected double dP, dQ;
    protected boolean zyklisch;

    /**
     * Konstruktor.
     * Erzeuge eine Kachel an der angegebenen Position mit der
     * angegebenen Faerbung. Die Ausgabeleinwand hat dabei die
     * angegebene Hoehe und Breite, den angegebenen Fluchtpunkt
     * und die angegebene Raumverzerrung hat.
     */
    public Kachel (int x, int y, Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
    {
        bild = new Vieleck ();
        
        positionX = x;
        positionY = y;
        
        faerbung = dieseFaerbung;
        
        hellsichtweite = weite;
    }
    
    /**
     * Mache die Kachel sichtbar.
     */
    public void macheSichtbar ()
    {
        bild.macheSichtbar ();
    }
    
    /**
     * Mache die Kachel unsichtbar.
     */
    public void macheUnsichtbar ()
    {
        bild.macheUnsichtbar ();
    }
    
    /**
     * Aendere die Leinwanddaten.
     * Die Kachel muss entsprechend den neuen Leinwanddaten geformt
     * und gefaerbt werden.
     */
    public abstract void aendereLeinwanddaten (Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite);
     
    /**
     * Nenne die x-Koordinate der Kachel.
     */
    public int gibX ()
    {
        return positionX;
    }
    
    /**
     * Nenne die y-Koordinate der Kachel.
     */
    public int gibY ()
    {
        return positionY;
    }
    
    /**
     * Nenne den Abstand vom Betrachter (in Summenmetrik).
     */
    public int gibAbstand ()
    {
        return abstand;
    }
    
    /**
     * Nenne den euklidischenAbstand vom Betrachter.
     */
    public double gibEuklidischenAbstand ()
    {
        return euklidischerAbstand;
    }
    
    /**
     * Pruefe, ob die Kachel eine Wandkachel ist.
     */
    public abstract boolean istWandkachel ();
    
    /**
     * Pruefe, ob die Kachel eine Bodenmarke ist.
     */
    public boolean istMarke ()
    {
        /*
         * Die meisten Kacheln sind keine Marken.
         */
        return false;
    }
    
    /**
     * Setze fernen Farbton entsprechend den Lichtverhaeltnissen.
     *
     * @param farbe       Farbe.
     * @param distanz     Euklidische Entfernung zum Ausgangspunkt.
     * @return            Variierte Farbe.
     */
    private Color erzeugeFernenFarbton (Color farbe, double distanz)
    {
        /*
         * Multipliziere zunaechst alle Farben mit dem
         * Helligkeitsfaktor.
         */
        double faktor = Math.pow ((2 * hellsichtweite / distanz), 2);
        
        int rot   = (int) Math.round (farbe.getRed () * faktor);
        int gruen = (int) Math.round (farbe.getGreen () * faktor);
        int blau  = (int) Math.round (farbe.getBlue () * faktor);
        
        /*
         * Sieh dann in der unmittelbaren Naehe des Betrachters eine
         * Weissverschiebung vor.
         */
        int betrag = (int) Math.round (Math.sqrt (Math.pow (1 + farbe.getRed (), 2)
                                                  + Math.pow (1 + farbe.getGreen (), 2)
                                                  + Math.pow (1 + farbe.getBlue (), 2)));
        
        int ueberschuss =Math.abs (rot - 255) + Math.abs (gruen - 255) + Math.abs (blau - 255);
        
        return new Color (Math.min (rot + ueberschuss / betrag, 255),
                          Math.min (gruen + ueberschuss / betrag, 255),
                          Math.min (blau + ueberschuss / betrag, 255));
    }
    
    /**
     * Erzeuge eine linear-orthogonale Naeherung fuer die
     * euklidische Faerbung des Bildes.
     */
    protected GradientPaint erzeugeBildfaerbung ()
    {
        return new GradientPaint (P, erzeugeFernenFarbton (faerbung, dP), Q, erzeugeFernenFarbton (faerbung, dQ), zyklisch);
    }
}
