import java.awt.Color;
import java.util.ArrayList;
import java.util.Collections;

/**
 * Eine dreidimensionale Darstellung des Labyrinths.
 * 
 * @author Bernd Gramlich 
 * @version 2013-02-06
 */
public class Labyrinth3DDarstellung implements PositionsBeobachter, LeinwandBeobachter
{
    /*
     * Die Codezeichen, die wir im Labyrinth erwarten.
     */
    private final char WAND             = '#';
    private final char LEERRAUM         = '.';
    private final char EINGANG          = '@';
    private final char SCHATZ           = '$';
    private final char MARKEVORWAERTS   = '^';
    private final char MARKELINKS       = '<';
    private final char MARKERUECKWAERTS = 'v';
    private final char MARKERECHTS      = '>';
    
    /*
     * Das Labyrinthmodell, das mich mit Labyrinthdaten
     * versorgt.
     */
    LabyrinthModel model;
    
    /*
     * Die Breite und Hoehe der Leinwand.
     */
    private int breite;
    private int hoehe;
    
    /*
     * Die relativen Koordinaten des Fluchtpunkts. Bei
     * quadratischer Leinwand bedeutet:
     * 
     * xfr == 0: Fluchtpunkt liegt am linken Bildrand.
     * xfr == 1: Fluchtpunkt liegt am rechten Bildrand.
     * yfr == 0: Fluchtpunkt liegt am oberen Bildrand.
     * yfr == 1: Fluchtpunkt liegt am unteren Bildrand.
     */
    private double xfr;
    private double yfr;
    
    /*
     * Die absoluten Koordinaten des Fluchtpunkts auf der
     * Leinwand.
     */
    private int xf;
    private int yf;
    
    /*
     * Der Verkuerzungsfaktor fuer die erste halbe Wandkachel,
     * die nach hinten fuehrt. (Die erste halbe Wandkachel hat
     * den mu-fachen Abstand von den Bildschirmkanten zum
     * Fluchtpunkt als Tiefe. Sinnvolle Werte fuer mu liegen
     * daher zwischen 0 und 1; optisch schoene Werte sind
     * meiner Meinung nach zwischen 0.25 und 0.3.)
     */
    private double mu;
    
    /*
     * Die Entfernung, gemessen in Kacheln, in der man
     * die Farben deutlich erkennt. (Bei geringerer
     * Entfernung werden sie heller, bei groesserer
     * dunkler, bis sie schliesslich in Schwarz uebergehen.)
     */
    private double hellsichtweite;
    
    /*
     * Die Grundfarben des Labyrinths.
     */
    private Color wandfarbe;
    private Color bodenfarbe;
    private Color deckenfarbe;
    private Color eingangsfarbe;
    private Color schatzfarbe;
    private Color markenfarbe;
    
    /*
     * Die Wandelemente.
     */
    private Kachel[][] boden;
    private Kachel[][] decke;
    private Kachel[][] wand;
    private Marke[][]  marke;
    
    /*
     * Ist an der betreffenden Stelle tatsaechlich eine Wand?
     */
    private char[][] ansicht;
    
    /*
     * Die maximale Blicktiefe, gemessen in Kacheln.
     */
    private int maxtiefe;
    
    /*
     * Die Liste aller zum Zeichnen vorgesehenen Kacheln.
     */
    private ArrayList<Kachel> vorrat;
    
    /*
     * Die Liste der tatsaechlich zu zeichnenden (also momentan
     * nicht verdeckten) Kacheln.
     */
    private ArrayList<Kachel> auswahl;
    
    /*
     * Die Liste der zuletzt gezeichneten Kacheln.
     */
    private ArrayList<Kachel> alteAuswahl;

    /**
     * Konstruktor.
     * Erzeuge eine Darstellung des LabyrinthModels m mit Breite b.
     */
    public Labyrinth3DDarstellung (LabyrinthModel m, int b)
    {
        model = m;
        
        breite = b;
        hoehe  = b;
        
        Leinwand.erzeugeLeinwand ("Labyrinth3DDarstellung", breite, hoehe, Color.black);
        Leinwand.gibLeinwand ().speichereAlsBeobachter (this);
        
        maxtiefe = model.gibSeitenlaenge ();
        
        boden = new Kachel[4*maxtiefe + 1][2*maxtiefe + 1];
        decke = new Kachel[4*maxtiefe + 1][2*maxtiefe + 1];
        wand  = new Kachel[4*maxtiefe + 1][2*maxtiefe + 1];
        marke = new Marke[4*maxtiefe + 1][2*maxtiefe + 1];
        
        vorrat      = new ArrayList<Kachel> ();
        auswahl     = new ArrayList<Kachel> ();
        alteAuswahl = new ArrayList<Kachel> ();
        
        wandfarbe     = Color.orange;
        bodenfarbe    = Color.lightGray;
        deckenfarbe   = Color.gray;
        eingangsfarbe = Color.cyan;
        schatzfarbe   = Color.yellow;
        markenfarbe   = Color.red;
        
        for (int x = - 2*maxtiefe; x <= 2*maxtiefe; x += 2)
        {
            for (int y = 0; y <= 2*maxtiefe; y += 2)
            {
                boden[index (x)][index (y)] = new Bodenkachel  (x, y, bodenfarbe,  breite, hoehe, xf, yf, mu, hellsichtweite);
                vorrat.add (boden[index (x)][index (y)]);
                decke[index (x)][index (y)] = new Deckenkachel (x, y, deckenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                vorrat.add (decke[index (x)][index (y)]);
                marke[index (x)][index (y)]   = new Marke (x, y, markenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                vorrat.add (marke[index (x)][index (y)]);
            }
        }
        
        for (int x = 1; x <= 2*maxtiefe - 1; x += 2)
        {
            for (int y = 0; y <= 2*maxtiefe; y += 2)
            {
                wand[index (-x)][index (y)] = new LinkeWandkachel  (-x, y, wandfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                vorrat.add (wand[index (-x)][index (y)]);
                wand[index (x)][index (y)] = new RechteWandkachel ( x, y, wandfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                vorrat.add (wand[index  (x)][index (y)]);
            }
        }
        
        for (int x = -2*maxtiefe; x <= 2*maxtiefe; x += 2)
        {
            for (int y = 1; y <= 2*maxtiefe - 1; y += 2)
            {
                wand[index (x)][index (y)] = new HintereWandkachel (x, y, wandfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                vorrat.add (wand[index (x)][index (y)]);
            }
        }
        
        Collections.sort (vorrat, new Kachelvergleicher ());
        
        ansicht = new char[4*maxtiefe + 1][4*maxtiefe + 1];
        
        setzeFluchtpunkt (0.5, 0.5);
        setzeHalblaenge (0.28);
        setzeHellsichtweite (1);
        
        model.speichereAlsBeobachter (this);
    }
    
    /**
     * Setze die relativen Koordinaten des Fluchtpunkts.
     * 
     * xFluchtpunkt == 0: Fluchtpunkt liegt am linken Bildrand.
     * xFluchtpunkt == 1: Fluchtpunkt liegt am rechten Bildrand.
     * yFluchtpunkt == 0: Fluchtpunkt liegt am oberen Bildrand.
     * yFluchtpunkt == 1: Fluchtpunkt liegt am unteren Bildrand.
     */
    public void setzeFluchtpunkt (double xFluchtpunkt, double yFluchtpunkt)
    {
        xfr = xFluchtpunkt;
        yfr = yFluchtpunkt;
        
        xf = (int) Math.round (xfr * breite);
        yf = (int) Math.round (yfr * hoehe);
        
        faerbeNeu ();
        zeichneNeu (); 
    }
    
    /**
     * Setze die relative raeumliche Laenge fuer die halbe Wand, neben
     * der wir stehen.
     * 
     * Werte kleiner als 0.1 werden zu 0.1, Werte groesser als 0.5 zu
     * 0.5 umgewandelt.
     */
    public void setzeHalblaenge (double halblaenge)
    {
        if (halblaenge < .1)
        {
            mu = .1;
        }
        else if (halblaenge > .5)
        {
            mu = .5;
        }
        else
        {
            mu = halblaenge;
        }
        
        faerbeNeu ();
        zeichneNeu ();
    }
    
    /**
     * Setze die Hellsichtweite.
     */
    public void setzeHellsichtweite (double weite)
    {
        if (weite < .1)
        {
            hellsichtweite = .1;
        }
        else if (weite > maxtiefe)
        {
            hellsichtweite = maxtiefe;
        }
        else
        {
            hellsichtweite = weite;
        }
        
        faerbeNeu ();
        zeichneNeu ();
    }

    /**
     * Setze die Wandfarbe auf den angegebenen Rot-Gruen-Blau-Wert. 
     */
    public void setzeWandfarbe (int rot, int gruen, int blau)
    {
        wandfarbe = new Color (rot, gruen, blau);
        
        faerbeNeu ();
        zeichneNeu ();
    }

    /**
     * Setze die Deckenfarbe auf den angegebenen Rot-Gruen-Blau-Wert. 
     */
    public void setzeDeckenfarbe (int rot, int gruen, int blau)
    {
        deckenfarbe = new Color (rot, gruen, blau);
        
        faerbeNeu ();
        zeichneNeu ();
    }

    /**
     * Setze die Bodenfarbe auf den angegebenen Rot-Gruen-Blau-Wert. 
     */
    public void setzeBodenfarbe (int rot, int gruen, int blau)
    {
        bodenfarbe = new Color (rot, gruen, blau);
        
        faerbeNeu ();
        zeichneNeu ();
    }

    /**
     * Setze die Eingangsfarbe auf den angegebenen Rot-Gruen-Blau-Wert. 
     */
    public void setzeEingangsfarbe (int rot, int gruen, int blau)
    {
        eingangsfarbe = new Color (rot, gruen, blau);
        
        faerbeNeu ();
        zeichneNeu ();
    }

    /**
     * Setze die Schatzfarbe auf den angegebenen Rot-Gruen-Blau-Wert. 
     */
    public void setzeSchatzfarbe (int rot, int gruen, int blau)
    {
        schatzfarbe = new Color (rot, gruen, blau);
        
        faerbeNeu ();
        zeichneNeu ();
    }

    /**
     * Setze die Markenfarbe auf den angegebenen Rot-Gruen-Blau-Wert. 
     */
    public void setzeMarkenfarbe (int rot, int gruen, int blau)
    {
        markenfarbe = new Color (rot, gruen, blau);
        
        faerbeNeu ();
        zeichneNeu ();
    }
    
    /**
     * Rechne eine Koordinate in einen Feldindex um.
     */
    private int index (int k)
    {
        return 2*maxtiefe - k;
    }
    
    /**
     * Faerbe die Kacheln neu.
     */
    private void faerbeNeu ()
    {
        if (null != ansicht)
        {
            for (int x = 1; x <= 2*maxtiefe - 1; x += 2)
            {
                for (int y = 0; y <= 2*maxtiefe; y += 2)
                {
                    wand[index (-x)][index (y)].aendereLeinwanddaten (wandfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                    wand[index ( x)][index (y)].aendereLeinwanddaten (wandfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                }
            }
        
            for (int x = -2*maxtiefe; x <= 2*maxtiefe; x += 2)
            {
                for (int y = 1; y <= 2*maxtiefe - 1; y += 2)
                {
                    wand[index (x)][index (y)].aendereLeinwanddaten (wandfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                }
            }
        
            for (int x = - 2*maxtiefe; x <= 2*maxtiefe; x += 2)
            {
                for (int y = 0; y <= 2*maxtiefe; y += 2)
                {
                    switch (ansicht[index (x)][index (y)])
                    {
                        case EINGANG:
                        boden[index (x)][index (y)].aendereLeinwanddaten (eingangsfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (eingangsfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        break;
                        case SCHATZ:
                        boden[index (x)][index (y)].aendereLeinwanddaten (schatzfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (schatzfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        break;
                        case MARKEVORWAERTS:
                        boden[index (x)][index (y)].aendereLeinwanddaten (bodenfarbe,  breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (deckenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        marke[index (x)][index (y)].aendereLeinwanddaten (markenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite, 0, 1);
                        break;
                        case MARKELINKS:
                        boden[index (x)][index (y)].aendereLeinwanddaten (bodenfarbe,  breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (deckenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        marke[index (x)][index (y)].aendereLeinwanddaten (markenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite, -1, 0);
                        break;
                        case MARKERUECKWAERTS:
                        boden[index (x)][index (y)].aendereLeinwanddaten (bodenfarbe,  breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (deckenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        marke[index (x)][index (y)].aendereLeinwanddaten (markenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite, 0, -1);
                        break;
                        case MARKERECHTS:
                        boden[index (x)][index (y)].aendereLeinwanddaten (bodenfarbe,  breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (deckenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        marke[index (x)][index (y)].aendereLeinwanddaten (markenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite, 1, 0);
                        break;
                        default:
                        boden[index (x)][index (y)].aendereLeinwanddaten (bodenfarbe,  breite, hoehe, xf, yf, mu, hellsichtweite);
                        decke[index (x)][index (y)].aendereLeinwanddaten (deckenfarbe, breite, hoehe, xf, yf, mu, hellsichtweite);
                        break;
                    }
                }
            }
        }
    }
    
    /**
     * Zeichne das Labyrinth neu.
     */
    public void zeichneNeu ()
    {
        if (null != ansicht)
        {
            /*
             * Mache alle Kacheln unsichtbar, die bisher gezeichnet
             * waren.
             */
            for (Kachel dieseKachel: alteAuswahl)
            {
                dieseKachel.macheUnsichtbar ();
            }
        
            /*
             * Mache alle Kacheln sichtbar, die gezeichnet werden
             * sollen. Beginne dabei mit der am weitesten
             * entfernten Kachel, damit die Kacheln korrekt
             * ueberlagert werden.
             */
            for (int i = auswahl.size () - 1; i >= 0; i -= 1)
            {
                auswahl.get (i).macheSichtbar ();
            }
        
            Leinwand.gibLeinwand ().erneutZeichnen ();
        }
    }
    
    /**
     * Ermittle die Zeichenelemente.
     * Speichere in der Liste "auswahl" nur diejenigen Kacheln,
     * die tatsaechlich gezeichnet werden sollen.
     * (In einer spaeteren Version sollen hier zur
     * Geschwindigkeitssteigerung nur diejenigen Kacheln
     * gespeichert werden, die nicht durch andere Kacheln
     * verdeckt sind.)
     */
    public void ermittleZeichenelemente ()
    {
        /*
         * Wir speichern die bisherigen Zeichenelemente.
         */
        alteAuswahl.clear ();
        
        for (Kachel dieseKachel: auswahl)
        {
            alteAuswahl.add (dieseKachel);
        }
        
        /*
         * Wir fangen ohne Zeichenelemente an.
         */
        auswahl.clear ();
            
        /*
         * Durchlaufe alle Kacheln des Vorrats.
         */
        for (Kachel dieseKachel: vorrat)
        {
            /*
             * Wenn diese Kachel gezeichnet werden soll ...
             */
            if (dieseKachel.istWandkachel ())
            {
                /*
                 * Merke eine Wandkachel nur zum Zeichnen vor, wenn an ihrer
                 * Position eine Wand gezeichnet werden soll.
                 */
                if (WAND == ansicht[index(dieseKachel.gibX ())][index(dieseKachel.gibY ())])
                {
                    auswahl.add (dieseKachel);
                }
            }
            else if (dieseKachel.istMarke ())
            {
                /*
                 * Merke eine Marke nur zum Zeichnen vor, wenn an ihrer
                 * Position eine Marke gezeichnet werden soll.
                 */
                switch (ansicht[index(dieseKachel.gibX ())][index(dieseKachel.gibY ())])
                {
                    case MARKEVORWAERTS:
                    case MARKELINKS:
                    case MARKERUECKWAERTS:
                    case MARKERECHTS:
                    auswahl.add (dieseKachel);
                }
            }
            else
            {
                /*
                 * Merke eine Boden- oder Deckenkachel immer zum
                 * Zeichnen vor.
                 */
                auswahl.add (dieseKachel);
            }
        }
    }
    
    /**
     * Aktualisiere Ansicht des Labyrinths.
     */
    public void aktualisiereLabyrinthansicht ()
    {
        ansicht = model.gibLabyrinthansicht ();
        
        ermittleZeichenelemente ();
        faerbeNeu ();
        zeichneNeu ();
    }
    
    /**
     * Aktualisiere die Masse der Leinwand, also die
     * Breite und Hoehe.
     */
    public void aktualisiereLeinwandmasse ()
    {
        breite = Leinwand.gibLeinwand ().gibBreite ();
        hoehe  = Leinwand.gibLeinwand ().gibHoehe ();
        
        ermittleZeichenelemente ();
        setzeFluchtpunkt (xfr, yfr);
    }
}
